/*
 * Copyright (c) 1999-2023 Logitech, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuLogitechTapPlugin, fu_logitech_tap_plugin, FU, LOGITECH_TAP_PLUGIN, FuPlugin)
