/*
 * Copyright (C) 2024 Algoltek, Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuAlgoltekUsbPlugin, fu_algoltek_usb_plugin, FU, ALGOLTEK_USB_PLUGIN, FuPlugin)
