/*
 * Copyright 2024 Dell Technologies
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later OR MIT
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuDellKestrelPlugin, fu_dell_kestrel_plugin, FU, DELL_KESTREL_PLUGIN, FuPlugin)
