/*
 * Copyright (C) 2021 Texas Instruments Incorporated
 * Copyright (C) 2022 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+ OR MIT
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuTiTps6598xPlugin, fu_ti_tps6598x_plugin, FU, TI_TPS6598X_PLUGIN, FuPlugin)
