/*
 * Copyright 2023 GN Audio A/S
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuJabraFilePlugin, fu_jabra_file_plugin, FU, JABRA_FILE_PLUGIN, FuPlugin)
