/*
 * Copyright (C) 2023 GN Audio A/S
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuJabraGnpPlugin, fu_jabra_gnp_plugin, FU, JABRA_GNP_PLUGIN, FuPlugin)
