/*
 * Copyright (C) 2023 Goodix.inc <xulinkun@goodix.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuGoodixtpPlugin, fu_goodixtp_plugin, FU, GOODIXTP_PLUGIN, FuPlugin)
