/*
 * Copyright (C) 2021 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

#define FU_TYPE_NORDIC_HID_ARCHIVE (fu_nordic_hid_archive_get_type())
G_DECLARE_FINAL_TYPE(FuNordicHidArchive,
		     fu_nordic_hid_archive,
		     FU,
		     NORDIC_HID_ARCHIVE,
		     FuArchiveFirmware)
