/*
 * Copyright (C) 2023 Dell Technologies
 * Copyright (C) 2023 Mediatek Inc.
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuMediatekScalerPlugin,
		     fu_mediatek_scaler_plugin,
		     FU,
		     MEDIATEK_SCALER_PLUGIN,
		     FuPlugin)
