/*
 * Copyright (C) 2023 Denis Pynkin <denis.pynkin@collabora.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuGenesysGl32xxPlugin,
		     fu_genesys_gl32xx_plugin,
		     FU,
		     GENESYS_GL32XX_PLUGIN,
		     FuPlugin)
