/*
 * Copyright 2024 Richard hughes <Richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuSynapticsVmm9Plugin,
		     fu_synaptics_vmm9_plugin,
		     FU,
		     SYNAPTICS_VMM9_PLUGIN,
		     FuPlugin)
