/*
 * Copyright (C) 2020 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <glib-object.h>

#define FU_HAILUCK_REPORT_ID_SHORT 0x05
#define FU_HAILUCK_REPORT_ID_LONG  0x06
