/*
 * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
 * Copyright (C) 2019 Mario Limonciello <mario.limonciello@dell.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#include <gio/gio.h>

#include "fu-volume.h"

FuVolume *
fu_volume_new_from_mount_path(const gchar *mount_path) G_GNUC_NON_NULL(1);
