/*
 * Copyright 2024 Mario Limonciello <superm1@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#pragma once

#include <fwupdplugin.h>

G_DECLARE_FINAL_TYPE(FuLegionHid2Plugin, fu_legion_hid2_plugin, FU, LEGION_HID2_PLUGIN, FuPlugin)
