// Copyright 2024 Richard Hughes <richard@hughsie.com>
// SPDX-License-Identifier: LGPL-2.1-or-later

#[derive(ToString)]
enum FuColorhugError {
    None,
    UnknownCmd,
    WrongUnlockCode,
    NotImplemented,
    UnderflowSensor,
    NoSerial,
    Watchdog,
    InvalidAddress,
    InvalidLength,
    InvalidChecksum,
    InvalidValue,
    UnknownCmdForBootloader,
    NoCalibration,
    OverflowMultiply,
    OverflowAddition,
    OverflowSensor,
    OverflowStack,
    DeviceDeactivated,
    IncompleteRequest,
    SelfTestSensor,
    SelfTestRed,
    SelfTestGreen,
    SelfTestBlue,
    SelfTestColorSelect,
    SelfTestMultiplier,
    InvalidCalibration,
    SramFailed,
    OutOfMemory,
    SelfTestTemperature,
    SelfTestI2c,
    SelfTestAdcVdd,
    SelfTestAdcVss,
    SelfTestAdcVref,
    I2cTargetAddress,
    I2cTargetConfig,
    SelfTestEeprom,
}
