/*
 * Copyright (c) 1999-2023 Logitech, Inc.
 * All Rights Reserved
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

syntax = "proto3";

package logi.device.proto;

option java_package = "com.logitech.vc.proto";

/**
 * Request for certificate chain
 * This is to be included in UsbMsg

 * EXPECTED RESPONSE
 * GetCertificateChainResponse
 */
message GetCertificateChainRequest
{
	/**
	 * attestation challenge
	 */
	string attestation = 1;

	/**
	 * time to live
	 */
	int32 ttl = 2;
}

/**
 * Get certificate chain response
 */
message GetCertificateChainResponse
{
	/**
	 * array of certs
	 */
	repeated string certchain = 1;
}
